import whendo.core.actions.logic_action as logic_x
from whendo.core.action import Action
import pytest


def negate(action: Action):
    return logic_x.Not(operand=action)


def is_exception(result):
    return isinstance(result, Exception)


def computes_exception(thunk):
    try:
        result = thunk()
    except Exception as exception:
        result = exception
    return isinstance(result, Exception)


# =================== tests ==================?


def test_success_1():
    assert not computes_exception(logic_x.Success().execute)


def test_exception_1():
    assert computes_exception(logic_x.Failure().execute)


def test_not_1():
    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            raise Exception()

    action1 = Action1()
    action2 = negate(action1)
    action3 = negate(action2)
    assert computes_exception(action1.execute)
    assert not computes_exception(action2.execute)
    assert computes_exception(action3.execute)


def test_not_2():
    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            return True

    action1 = Action1()
    action2 = negate(action1)
    action3 = negate(action2)
    assert not computes_exception(action1.execute)
    assert computes_exception(action2.execute)
    assert not computes_exception(action3.execute)


def test_list_action_all_1():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action4(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    all_action = logic_x.All(
        op_mode=logic_x.ListOpMode.ALL,
        action_list=[Action1(), Action2(), Action3(), Action4()],
    )
    result = all_action.execute()
    assert dictionary["value"] == Action4


def test_list_action_or_1():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action4(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    or_action = logic_x.Or(
        action_list=[negate(Action1()), Action2(), Action3(), Action4()],
    )
    result = or_action.execute()
    assert dictionary["value"] == Action2


def test_list_action_and_1():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action4(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    and_action = logic_x.And(
        action_list=[Action1(), Action2(), Action3(), Action4()],
    )
    result = and_action.execute()
    assert dictionary["value"] == Action4


def test_list_action_and_2():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    class Action4(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__

    and_action = logic_x.And(
        action_list=[
            negate(Action1()),
            negate(Action2()),
            negate(Action3()),
            negate(Action4()),
        ]
    )
    result = and_action.execute()
    assert dictionary["value"] == Action1


def test_list_action_and_3():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            return Exception()

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            return Exception()

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            return Exception()

    class Action4(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            return Exception()

    and_action = logic_x.And(
        action_list=[Action1(), Action2(), Action3(), Action4()],
        exception_on_no_success=True,
    )
    assert computes_exception(and_action.execute)


def test_if_else_action_or_2():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__
            return True

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__
            return True

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__
            return True

    if_else_action = logic_x.IfElse(
        test_action=Action1(),
        else_action=Action2(),
        if_action=Action3(),
        exception_on_no_success=False,
    )
    result = if_else_action.execute()
    assert dictionary["value"] == Action3


def test_if_else_action_else_1():
    dictionary = {"value": None}

    class Action1(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            return Exception()

    class Action2(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__
            return True

    class Action3(Action):
        def execute(self, tag: str = None, scheduler_info: dict = None):
            dictionary["value"] = self.__class__
            return True

    if_else_action = logic_x.IfElse(
        test_action=Action1(),
        else_action=Action2(),
        if_action=Action3(),
        exception_on_no_success=False,
    )
    result = if_else_action.execute()
    assert dictionary["value"] == Action2



from typing import Dict, Any, Optional
import os
import logging
from whendo.core.util import PP, Dirs, Now, SystemInfo
from whendo.core.action import Action

logger = logging.getLogger(__name__)


class FileHeartbeat(Action):
    """
    This class is primarily for testing and heartbeat purposes.
    """

    file: str
    relative_to_output_dir: bool = True
    xtra: Optional[Dict[str, Any]] = None

    def description(self):
        return f"This action appends various status information to ({self.file})."

    def execute(self, tag: str = None, scheduler_info: Dict[str, Any] = None):
        payload = self.build_payload(self.info(), scheduler_info)
        payload["tag"] = tag if tag else "N/A"
        if self.xtra:
            payload[
                "xtra"
            ] = (
                self.xtra
            )  # dictionaries are sorted by key. Nice to have extra information at the bottom.
        file = (
            os.path.join(Dirs.output_dir(), self.file)
            if self.relative_to_output_dir
            else self.file
        )
        with open(file, "a") as outfile:
            PP.pprint(payload, stream=outfile)
            outfile.write("\n")
        return {"outcome": "file appended", "action": self.info()}

    def set_xtra(self, xtra: Dict[str, Any] = None):
        self.xtra = xtra
        return self

    def get_xtra(self):
        return self.xtra

    def build_payload(
        self, action_info: Dict[str, Any], scheduler_info: Dict[str, Any] = None
    ):
        payload = {}
        payload.update({"action_info": action_info})
        if scheduler_info:
            payload.update({"scheduler_info": scheduler_info})
        payload.update(self.action_host())
        payload.update(self.action_time())
        return payload

    def action_host(self):
        return {"action_host": SystemInfo.get()["host"]}

    def action_time(self):
        return {"action_time": Now.s()}



import pytest
import time
from datetime import timedelta
import whendo.core.util as util
from whendo.core.action import Action
from whendo.core.schedulers.cont_scheduler import Immediately, Timely
from whendo.core.dispatcher import Dispatcher
from whendo.core.program import Program
from whendo.core.continuous import Continuous

pause = 2


def test_schedule_action(friends):
    """
    Tests Dispatcher and Continuous objects running a scheduled action.
    """
    dispatcher, scheduler, action = friends()

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")

    assert dispatcher.get_scheduled_action_count() == 1

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert action.flea_count > 0


def test_unschedule_action(friends):
    """
    Tests unscheduling an action.
    """
    dispatcher, scheduler, action = friends()

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    assert dispatcher.job_count() == 1
    dispatcher.unschedule_action("foo")
    assert dispatcher.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == dispatcher.job_count()


def test_reschedule_action(friends):
    """
    Tests unscheduling and then rescheduling an action.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    dispatcher.reschedule_action("foo")

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert action.flea_count > 0


def test_reschedule_action_2(friends, tmp_path):
    """
    Tests unscheduling and then rescheduling an action.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    stored_action = dispatcher.get_action("foo")
    stored_action.flea_count = 100000
    dispatcher.set_action("foo", stored_action)
    dispatcher.reschedule_action("foo")
    retrieved_action = dispatcher.get_action("foo")

    dispatcher.run_jobs()
    time.sleep(pause)
    dispatcher.stop_jobs()
    dispatcher.clear_jobs()

    assert retrieved_action.flea_count > 100000


def test_unschedule_scheduler(friends):
    """
    Tests unscheduling a scheduler.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")

    assert dispatcher.job_count() == 1

    dispatcher.unschedule_scheduler("bar")

    assert dispatcher.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == dispatcher.job_count()

    # make sure that bar and foo remain
    assert dispatcher.get_scheduler("bar")
    assert dispatcher.get_action("foo")


def test_reschedule_all(friends):
    """
    Tests unscheduling a scheduler.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    assert dispatcher.job_count() == 1

    dispatcher.clear_jobs()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo2", action)
    dispatcher.schedule_action("bar", "foo2")
    assert dispatcher.job_count() == 1

    dispatcher.reschedule_all_schedulers()
    assert dispatcher.job_count() == 2

    assert dispatcher.get_scheduled_action_count() == dispatcher.job_count()


def test_clear_dispatcher(friends):
    """
    Tests clearing a dispatcher.
    """
    dispatcher, scheduler, action = friends()
    assert dispatcher.job_count() == 0

    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action("bar", "foo")
    assert dispatcher.job_count() == 1

    dispatcher.clear_all()
    assert dispatcher.job_count() == 0
    assert dispatcher.get_scheduled_action_count() == dispatcher.job_count()

    # make sure that bar and foo are Gone
    assert dispatcher.get_scheduler("bar") is None
    assert dispatcher.get_action("foo") is None


def test_scheduled_action_count(friends):
    """
    Tests scheduled action count
    """
    # original
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action(action_name="foo", scheduler_name="bar")
    assert 1 == dispatcher.get_scheduled_action_count()
    assert 1 == dispatcher.job_count()


def test_jobs_are_running(friends):
    dispatcher, scheduler, action = friends()
    try:
        dispatcher.run_jobs()
        assert dispatcher.jobs_are_running()
    finally:
        try:
            dispatcher.stop_jobs()
        except:
            pass


def test_jobs_are_not_running(friends):
    dispatcher, scheduler, action = friends()
    try:
        dispatcher.run_jobs()
        assert dispatcher.jobs_are_running()
        dispatcher.stop_jobs()
        assert not dispatcher.jobs_are_running()
    finally:
        try:
            dispatcher.stop_jobs()
        except:
            pass


def test_replace_dispatcher(friends):
    """
    Tests replacing a dispatcher
    """
    # original
    dispatcher, scheduler, action = friends()
    saved_dir = dispatcher.get_saved_dir()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action(action_name="foo", scheduler_name="bar")
    # replacement
    replacement = Dispatcher()  # no saved_dir
    replacement.add_action("flea", action)
    replacement.add_scheduler("bath", scheduler)
    replacement.schedule_action(action_name="flea", scheduler_name="bath")
    # do the business
    dispatcher.replace_all(replacement)
    # is everyone okay?
    assert not dispatcher.get_action("foo")
    assert not dispatcher.get_scheduler("bar")
    assert dispatcher.get_action("flea")
    assert dispatcher.get_scheduler("bath")
    assert {"bath"} == set(k for k in dispatcher.get_schedulers())
    assert {"flea"} == set(k for k in dispatcher.get_actions())
    assert {"bath"} == set(k for k in dispatcher.get_scheduled_actions())
    assert {"flea"} == set(dispatcher.get_scheduled_actions()["bath"])


def test_load_dispatcher(friends):
    """
    Tests loading a dispatcher
    """
    dispatcher, scheduler, action = friends()
    saved_dir = dispatcher.get_saved_dir()
    assert saved_dir is not None
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.schedule_action(action_name="foo", scheduler_name="bar")
    dispatcher2 = dispatcher.load_current()
    assert dispatcher2 is not None, f"saved_dir({saved_dir})"
    assert set(k for k in dispatcher.get_actions()) == set(
        k for k in dispatcher2.get_actions()
    )
    assert set(k for k in dispatcher.get_schedulers()) == set(
        k for k in dispatcher2.get_schedulers()
    )
    assert set(k for k in dispatcher.get_scheduled_actions()) == set(
        k for k in dispatcher2.get_scheduled_actions()
    )


def test_saved_dir_1(tmp_path):
    saved_dir = str(tmp_path)
    dispatcher = Dispatcher()
    dispatcher.set_saved_dir(saved_dir=saved_dir)
    assert dispatcher.get_saved_dir() == saved_dir


def test_saved_dir_2(tmp_path):
    saved_dir = str(tmp_path)
    dispatcher = Dispatcher(saved_dir=saved_dir)
    assert dispatcher.get_saved_dir() == saved_dir


def test_defer_action(friends):
    """
    Want to observe the scheduling move from deferred state to ready state.
    """
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)

    assert 0 == dispatcher.get_deferred_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    dispatcher.defer_action(
        scheduler_name="bar", action_name="foo", wait_until=util.Now.dt()
    )

    # deferred state -- can run jobs and actions will _not_ be executed
    assert 1 == dispatcher.get_deferred_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    time.sleep(6)  # the out-of-band job runs every five seconds

    # ready state -- can run jobs and actions will be executed
    assert 0 == dispatcher.get_deferred_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()


def test_expire_action(friends):
    """
    Want to observe the scheduling move from deferred state to ready state.
    """
    dispatcher, scheduler, action = friends()
    dispatcher.add_action("foo", action)
    dispatcher.add_scheduler("bar", scheduler)

    assert 0 == dispatcher.get_expiring_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()

    dispatcher.schedule_action(scheduler_name="bar", action_name="foo")

    assert 0 == dispatcher.get_expiring_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()

    dispatcher.expire_action(
        scheduler_name="bar",
        action_name="foo",
        expire_on=util.Now.dt() + timedelta(seconds=1),
    )

    assert 1 == dispatcher.get_expiring_action_count()
    assert 1 == dispatcher.get_scheduled_action_count()

    time.sleep(6)  # the out-of-band job runs every 2-5 seconds

    assert 0 == dispatcher.get_expiring_action_count()
    assert 0 == dispatcher.get_scheduled_action_count()


def test_immediately(friends):
    """
    Want to observe that action get executed immediately and that schedulers_actions
    is not impacted.
    """

    dispatcher, scheduler, action = friends()

    class TestAction(Action):
        fleas: int = 0

        def execute(self, tag: str = None, scheduler_info: dict = None):
            self.fleas += 1
            return "BLING"

    test_action = TestAction()

    assert dispatcher.get_scheduled_action_count() == 0
    assert test_action.fleas == 0

    dispatcher.add_action("foo", test_action)
    dispatcher.add_scheduler("imm", Immediately())
    dispatcher.schedule_action(scheduler_name="imm", action_name="foo")

    assert dispatcher.get_scheduled_action_count() == 0
    assert test_action.fleas == 1


def test_program(friends):
    """
    Want to observe that a Program's actions are executed.
    """

    dispatcher, scheduler, action = friends()

    class TestAction1(Action):
        fleas: int = 0

        def execute(self, tag: str = None, scheduler_info: dict = None):
            self.fleas += 1
            return "BLING"

    class TestAction2(Action):
        fleas: int = 0

        def execute(self, tag: str = None, scheduler_info: dict = None):
            self.fleas += 1
            return "BLING"

    class TestAction3(Action):
        fleas: int = 0

        def execute(self, tag: str = None, scheduler_info: dict = None):
            self.fleas += 1
            return "BLING"

    action1 = TestAction1()
    action2 = TestAction2()
    action3 = TestAction3()

    dispatcher.add_action("foo1", action1)
    dispatcher.add_action("foo2", action2)
    dispatcher.add_action("foo3", action3)
    dispatcher.add_scheduler("bar", scheduler)
    dispatcher.add_scheduler("immediately", Immediately())

    program = Program().prologue("foo1").body_element("bar", "foo2").epilogue("foo3")
    dispatcher.add_program("baz", program)
    start = util.Now().dt()
    stop = start + timedelta(seconds=5)

    dispatcher.run_jobs()
    dispatcher.schedule_program("baz", start, stop)
    assert action1.fleas == 0
    time.sleep(1)
    assert action1.fleas == 1
    time.sleep(4)
    assert action2.fleas >= 2
    time.sleep(2)
    assert action3.fleas == 1


@pytest.fixture
def friends(tmp_path):
    """ returns a tuple of useful test objects """

    class FleaCount(Action):
        flea_count: int = 0

        def execute(self, tag: str = None, scheduler_info: dict = None):
            self.flea_count += 1

    def stuff():
        # want a fresh tuple from the fixture
        dispatcher = Dispatcher(saved_dir=str(tmp_path))
        dispatcher.set_continuous(Continuous())
        dispatcher.initialize()
        action = FleaCount()
        scheduler = Timely(interval=1)

        return dispatcher, scheduler, action

    return stuff



from pydantic import BaseModel, PrivateAttr
import requests
import logging
from typing import Optional
from whendo.core.action import Action
from whendo.core.scheduler import Scheduler
from whendo.core.resolver import (
    resolve_action,
    resolve_scheduler,
    resolve_file_pathe,
    resolve_program,
)
from whendo.core.util import FilePathe, DateTime, Http, DateTime2
from whendo.core.dispatcher import Dispatcher
from whendo.core.program import Program

default_host = "127.0.0.1"
default_port = 8000

logger = logging.getLogger(__name__)


class Client(BaseModel):
    host: str = default_host
    port: int = default_port
    client_used: bool = False

    # hidden private field
    _http: Http = PrivateAttr(default_factory=Http)

    def http(self):
        """
        Cache the Http object.
        """
        if not self.client_used:
            self._http = Http(host=self.host, post=self.port)
            self.client_used = True
        return self._http

    # /dispatcher
    def load_dispatcher(self):
        return Dispatcher.resolve(self.http().get(f"/dispatcher/load"))

    def save_dispatcher(self):
        return self.http().get("/dispatcher/save")

    def clear_dispatcher(self):
        return self.http().get("/dispatcher/clear")

    def load_dispatcher_from_name(self, name: str):
        return Dispatcher.resolve(self.http().get(f"/dispatcher/load_from_name/{name}"))

    def save_data_to_name(self, name: str):
        return self.http().get(f"/dispatcher/save_to_name/{name}")

    def get_saved_dir(self):
        return resolve_file_pathe(self.http().get("/dispatcher/saved_dir"))

    def set_saved_dir(self, saved_dir: FilePathe):
        return self.http().put("/dispatcher/saved_dir", saved_dir)

    def replace_dispatcher(self, replacement: Dispatcher):
        return self.http().put("/dispatcher/replace", replacement)

    def describe_all(self):
        return self.http().get("/dispatcher/describe_all")

    # /execution

    def execute_supplied_action(self, supplied_action: Action):
        return self.http().put(f"/execution", supplied_action)

    # /actions
    def get_action(self, action_name: str):
        return resolve_action(self.http().get(f"/actions/{action_name}"))

    def describe_action(self, action_name: str):
        return self.http().get(f"/actions/{action_name}/describe")

    def add_action(self, action_name: str, action: Action):
        return self.http().post(f"/actions/{action_name}", action)

    def set_action(self, action_name: str, action: Action):
        return self.http().put(f"/actions/{action_name}", action)

    def delete_action(self, action_name: str):
        return self.http().delete(f"/actions/{action_name}")

    def execute_action(self, action_name: str):
        return self.http().get(f"/actions/{action_name}/execute")

    def unschedule_action(self, action_name: str):
        return self.http().get(f"/actions/{action_name}/unschedule")

    def reschedule_action(self, action_name: str):
        return self.http().get(f"/actions/{action_name}/reschedule")

    # /schedulers

    def schedule_action(self, scheduler_name: str, action_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/actions/{action_name}")

    def get_scheduler(self, scheduler_name: str):
        return resolve_scheduler(self.http().get(f"/schedulers/{scheduler_name}"))

    def describe_scheduler(self, scheduler_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/describe")

    def add_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return self.http().post(f"/schedulers/{scheduler_name}", scheduler)

    def set_scheduler(self, scheduler_name: str, scheduler: Scheduler):
        return self.http().put(f"/schedulers/{scheduler_name}", scheduler)

    def delete_scheduler(self, scheduler_name: str):
        return self.http().delete(f"/schedulers/{scheduler_name}")

    def unschedule_scheduler(self, scheduler_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/unschedule")

    def unschedule_scheduler_action(self, scheduler_name: str, action_name: str):
        return self.http().get(
            f"/schedulers/{scheduler_name}/actions/{action_name}/unschedule"
        )

    def reschedule_all_schedulers(self):
        return self.http().get(f"/schedulers/reschedule_all")

    def execute_scheduled_actions(self, scheduler_name: str):
        return self.http().get(f"/schedulers/{scheduler_name}/execute")

    def scheduled_action_count(self):
        return self.http().get("/schedulers/action_count")

    # programs
    def get_program(self, program_name: str):
        return resolve_program(self.http().get(f"/programs/{program_name}"))

    def describe_program(self, program_name: str):
        return self.http().get(f"/programs/{program_name}/describe")

    def add_program(self, program_name: str, program: Program):
        return self.http().post(f"/programs/{program_name}", program)

    def set_program(self, program_name: str, program: Program):
        return self.http().put(f"/programs/{program_name}", program)

    def delete_program(self, program_name: str):
        return self.http().delete(f"/programs/{program_name}")

    def schedule_program(self, program_name: str, datetime2: DateTime2):
        return self.http().post(f"/programs/{program_name}/schedule", datetime2)

    # deferrals and expirations
    def defer_action(self, scheduler_name: str, action_name: str, wait_until: DateTime):
        return self.http().post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/defer", wait_until
        )

    def clear_deferred_actions(self):
        return self.http().get(f"/schedulers/clear_deferred_actions")

    def deferred_action_count(self):
        return self.http().get("/schedulers/deferred_action_count")

    def expire_action(self, scheduler_name: str, action_name: str, expire_on: DateTime):
        return self.http().post(
            f"/schedulers/{scheduler_name}/actions/{action_name}/expire", expire_on
        )

    def clear_expiring_actions(self):
        return self.http().get(f"/schedulers/clear_expiring_actions")

    def expiring_action_count(self):
        return self.http().get("/schedulers/expiring_action_count")

    # /jobs
    def run_jobs(self):
        return self.http().get(f"/jobs/run")

    def stop_jobs(self):
        return self.http().get(f"/jobs/stop")

    def jobs_are_running(self):
        return self.http().get(f"/jobs/are_running")

    def job_count(self):
        return self.http().get(f"/jobs/count")

    def clear_jobs(self):
        return self.http().get(f"/jobs/clear")



import time
from whendo.core.continuous import Continuous
from whendo.core.util import TimeUnit, PP
from whendo.core.action import Action

pause = 3


def test_timely_callable(tmp_path):
    """
    This test exercises the schedule_timely_callable method.
    """

    class Suite:
        def __init__(self):
            self.content = "foobarbaz"
            self.file = "text.txt"
            self.path = tmp_path / self.file

        def callable(self):
            with self.path.open(mode="a") as fid:
                fid.write(self.content)
                fid.write("\n")

        def gather(self):
            result = []
            with self.path.open(mode="r") as fid:
                result.append(fid.readlines())
            return result

        def run(self):
            continuous = Continuous()
            continuous.schedule_timely_callable("tag", self.callable)
            continuous.run_continuously()
            time.sleep(pause)
            continuous.stop_running_continuously()
            continuous.clear()

    suite = Suite()
    suite.run()
    accumulated_content = suite.gather()
    assert accumulated_content and len(accumulated_content) > 0


def test_random_callable(tmp_path):
    """
    This test exercises the schedule_timely_callable method.
    """

    class Suite:
        def __init__(self):
            self.content = "foobarbaz"
            self.file = "text.txt"
            self.path = tmp_path / self.file

        def callable(self):
            with self.path.open(mode="a") as fid:
                fid.write(self.content)
                fid.write("\n")

        def gather(self):
            result = []
            with self.path.open(mode="r") as fid:
                result.append(fid.readlines())
            return result

        def run(self):
            continuous = Continuous()
            continuous.schedule_random_callable(
                "tag", self.callable, time_unit=TimeUnit.second, low=1, high=3
            )
            continuous.run_continuously()
            time.sleep(4)
            continuous.stop_running_continuously()
            continuous.clear()

    suite = Suite()
    suite.run()
    accumulated_content = suite.gather()
    assert accumulated_content and len(accumulated_content) > 0


def test_file_action(tmp_path):
    """
    This test exercises the schedule_timely_callable method.
    """

    class FileAction(Action):
        def execute(self, tag=None, scheduler_info: dict = None):
            path = tmp_path / "test.txt"
            with path.open(mode="a") as fid:
                fid.write("blee\n")
            return {"outcome": "file appended"}

    class Suite:
        def __init__(self, action):
            self.action = action

        def gather(self):
            path = tmp_path / "test.txt"
            with path.open(mode="r") as fid:
                return sum(1 for line in fid)

        def run(self):
            continuous = Continuous()
            continuous.schedule_timely_callable("tag", self.action.execute)
            continuous.run_continuously()
            time.sleep(pause)
            continuous.stop_running_continuously()
            continuous.clear()

    suite = Suite(FileAction())
    suite.run()
    line_count = suite.gather()
    assert line_count and line_count >= 2, "no lines written to file"



from pydantic import BaseModel
import logging
from .util import object_info


logger = logging.getLogger(__name__)


class Action(BaseModel):
    """
    Actions get something done.
    """

    def description(self):
        return "This Action does nothing."

    def execute(self, tag: str = None, scheduler_info: dict = None):
        """
        This method typically attempts to do something useful and return something useful
        """
        return {"result": "something useful happened"}

    def info(self):
        return object_info(self)

    def flat(self):
        return self.json()



import time
import logging
from whendo.core.action import Action
import whendo.core.util as util


logger = logging.getLogger(__name__)


class SysInfo(Action):
    """
    Return system info
    """

    sys_info = "sys_info"

    def description(self):
        return f"This action returns system-level information."

    def execute(self, tag: str = None, scheduler_info: dict = None):
        return util.SystemInfo.get()


class Pause(Action):
    """
    Sleep for supplied int seconds. It's best to make sure that the sleep duration
    does not interfere with job execution (blocking the thread).
    """

    seconds: int = 1

    def description(self):
        return f"This action sleeps for {self.seconds} second{'s' if self.seconds > 1 else ''}."

    def execute(self, tag: str = None, scheduler_info: dict = None):
        return time.sleep(self.seconds)



import requests
from typing import Optional, Dict, Any
import logging
from whendo.core.action import Action
from whendo.core.util import Http


logger = logging.getLogger(__name__)


class SendPayload(Action):
    """
    This class sends a payload dictionary to a url.
    """

    url: str
    payload: Optional[Dict[str, Any]] = None

    def description(self):
        return f"This action sends the supplied dictionary payload to ({self.url})."

    def execute(self, tag: str = None, scheduler_info: dict = None):
        response = requests.get(self.url, self.payload)
        if response.status_code != requests.codes.ok:
            raise Exception(response)
        return response.json()


class ExecuteAction(Action):
    """
    Execute an action at host:port.
    """

    host: str
    port: int
    action_name: str

    def description(self):
        return f"This action executes ({self.action_name}) at host ({self.host}) and port ({self.port})."

    def execute(self, tag: str = None, scheduler_info: dict = None):
        return Http(host=self.host, port=self.port).get(
            f"/actions/{self.action_name}/execute"
        )



"""
That these modules are imported here ensures that Action and Scheduler resolution will use their entire
inheritance trees to resolve an instance from a dictionary. See ./resolver.py.
"""
import whendo.core.actions.file_action
import whendo.core.actions.http_action
import whendo.core.actions.logic_action
import whendo.core.actions.sys_action
import whendo.core.schedulers.cont_scheduler
